@@ -45,6 +45,7 @@ impl HolochainP2pHandler for StubNetwork {
         dna_hash: DnaHash,
         from_agent: AgentPubKey,
         request_validation_receipt: bool,
+        countersigning_session: bool,
         dht_hash: holo_hash::AnyDhtHash,
         ops: Vec<(holo_hash::DhtOpHash, holochain_types::dht_op::DhtOp)>,
         timeout_ms: Option<u64>,
@@ -103,6 +104,26 @@ impl HolochainP2pHandler for StubNetwork {
     ) -> HolochainP2pHandlerResult<()> {
         Err("stub".into())
     }
+    fn handle_new_integrated_data(&mut self, dna_hash: DnaHash) -> HolochainP2pHandlerResult<()> {
+        Err("stub".into())
+    }
+    fn handle_authority_for_hash(
+        &mut self,
+        dna_hash: DnaHash,
+        agent: AgentPubKey,
+        dht_hash: AnyDhtHash,
+    ) -> HolochainP2pHandlerResult<bool> {
+        Err("stub".into())
+    }
+    fn handle_countersigning_authority_response(
+        &mut self,
+        dna_hash: DnaHash,
+        from_agent: AgentPubKey,
+        agents: Vec<AgentPubKey>,
+        response: Vec<SignedHeader>,
+    ) -> HolochainP2pHandlerResult<()> {
+        Err("stub".into())
+    }
 }
 
 /// Spawn a stub network that doesn't respond to any messages.
@@ -329,7 +350,7 @@ mod tests {
         // this will fail because we can't reach any remote nodes
         // but, it still published locally, so our test will work
         let _ = p2p
-            .publish(dna, a1, true, header_hash, vec![], Some(200))
+            .publish(dna, a1, true, false, header_hash, vec![], Some(200))
             .await;
 
         assert_eq!(3, recv_count.load(std::sync::atomic::Ordering::SeqCst));
@@ -388,8 +409,8 @@ mod tests {
                     QueryAgentInfoSigned { respond, .. } => {
                         respond.r(Ok(async move { Ok(vec![]) }.boxed().into()));
                     }
-                    FetchOpHashesForConstraints { respond, .. } => {
-                        respond.r(Ok(async move { Ok(vec![]) }.boxed().into()));
+                    QueryOpHashes { respond, .. } => {
+                        respond.r(Ok(async move { Ok(None) }.boxed().into()));
                     }
                     evt => println!("unhandled: {:?}", evt),
                 }
